/*
 * strace.c
 * 
 * Copyright (c) 2000, BindView Corporation.
 *
 * See LICENSE file.
 *
 */

#include <tchar.h>
#include <windows.h>
#include <winnt.h>

#include <winioctl.h>
#include <stdio.h>

#include "syscalls.h"
#include "../driver/ioctlcmd.h"

extern BOOL LoadDeviceDriver (const TCHAR * Name,
                              const TCHAR * Path,
                              HANDLE * lphDevice);
BOOL UnloadDeviceDriver( const TCHAR * Name );
int enable_debug_priv (void);

const char *lookup_status (unsigned int status);


#define MagicFoo(call,args) #call,
char *syscall_names[] = {
#include "../driver/syscallx.h"
};
#undef MagicFoo

void
Usage ()
{
    fprintf (stderr, "Usage:\n\
    strace -p <pid>\n\
    strace <cmdline>\n");
    exit (1);
}


int
load_driver (HANDLE *pdev)
{
    char *p;
    char file_name[MAX_PATH+1];
    
    if (GetModuleFileName (GetModuleHandle (NULL),
                           file_name, sizeof (file_name)-1)) {
        p = strrchr (file_name, '\\');
        p++;
        strncpy (p, "strace.sys", sizeof (file_name) - (p - file_name));
        file_name[sizeof (file_name)-1] = 0;
    }
    return LoadDeviceDriver ("Strace", file_name, pdev);
}


int
main (int argc, char *argv[])
{
    HANDLE dev = 0;
    HANDLE hThread = 0;
    HANDLE hProcess = 0;
    DWORD bytes_out;
    unsigned int i;
    char Stats[MAX_DATA];
    DWORD StatsLen;
    DWORD pid = 0;
    char *cmd = NULL;
    int done = 0;
    unsigned int last_dir=1, last_seq=0;

    if (argc == 1) {
        Usage ();
    }
    if ((argc == 3) &&
        (strcmp (argv[1], "-p") == 0)) {
        pid = atoi (argv[2]);
    } else {
        cmd = GetCommandLine ();
        /*
         * skip strace and the space
         */
        cmd += strcspn (cmd, " ");
        cmd ++;
    }

    init_arg_info ();

    enable_debug_priv ();

    if (load_driver (&dev)) {
        if (cmd) {
            STARTUPINFO si;
            PROCESS_INFORMATION pi;

            memset (&si,0, sizeof (si));
            si.cb = sizeof (si);

            if (CreateProcess (NULL, cmd, NULL, NULL, FALSE,
                               CREATE_SUSPENDED, NULL, NULL, &si, &pi)) {
                hThread = pi.hThread;
                hProcess = pi.hProcess;
                pid = pi.dwProcessId;
            } else {
                printf ("CreateProcess failed: %x\n", GetLastError ());
                exit (1);
            }
        }
        
        if (!DeviceIoControl (dev, STRACE_hook,
                               &pid, 4, NULL, 0, &bytes_out, NULL))
        {
            printf ("ioctl - start failed 0x%x\n", GetLastError ());
            goto exit;
        }

        if (hThread) {
            ResumeThread (hThread);
            CloseHandle (hThread);
        }

        for ( ; ; ) {
            ENTRY *ent;
            DWORD bytes_used;
            do {
                StatsLen = 0;
                if (!DeviceIoControl (dev, STRACE_getdata, NULL, 0,
                                      &Stats, sizeof Stats,
                                      &StatsLen, NULL))
                {
                    printf ("getstats failed: 0x%x\n", GetLastError ());
                    goto exit;
                }
                bytes_used = 0;
                ent = (ENTRY *) Stats;
                while (bytes_used < StatsLen) {
                    int arg_bytes_used;

                    if (ent->direction == 0) {
                        if (last_dir == 0) {
                            printf ("\n");
                        }
                        printf ("%d %d %d %s (",
                                ent->seq,
                                ent->pid,
                                ent->tid,
                                syscall_names[ent->call_num]);

                        arg_bytes_used = 0;
                        for (i=0; i<all_arg_info[ent->call_num]->num; i++) {
                            if ((all_arg_info[ent->call_num]->args[i].dir == DIR_IN)
                                || (all_arg_info[ent->call_num]->args[i].dir == DIR_IN_OUT)) {

                                all_types[all_arg_info[ent->call_num]->args[i].type].copy_arg (ent, (DWORD) &arg_bytes_used);
                            }
                        }

                        printf ("... ");
                        last_dir = 0;
                        last_seq = ent->seq;
                    } else {
                        if ((last_seq != ent->seq)
                            && (last_dir != 1)) {
                            printf ("\n");
                        }
                        if (last_seq != ent->seq) {
                            printf ("%d %d %d %s ",
                                    ent->seq,
                                    ent->pid,
                                    ent->tid,
                                    syscall_names[ent->call_num]);
                        }
                        arg_bytes_used = 0;
                        for (i=0; i<all_arg_info[ent->call_num]->num; i++) {
                            if ((all_arg_info[ent->call_num]->args[i].dir == DIR_OUT)
                                || (all_arg_info[ent->call_num]->args[i].dir == DIR_IN_OUT)) {

                                all_types[all_arg_info[ent->call_num]->args[i].type].copy_arg (ent, (DWORD) &arg_bytes_used);
                            }
                        }
                        {
                            const char *st = lookup_status (ent->result);
                            if (st)
                                printf (") == %s\n", st);
                            else 
                                printf (") == 0x%x\n", ent->result);
                        }
                        last_dir = 1;
                        last_seq = ent->seq;
                    }


                    bytes_used += ENT_SIZE (ent);
                    ent = (ENTRY *)(((char *)ent)+ENT_SIZE (ent));
                }
            } while (StatsLen != 0);

            fflush (stdout);
            if (done)
                break;
            if (hProcess) {
                if (WaitForSingleObject (hProcess, 1000) == WAIT_OBJECT_0) {
                    done = 1;
                }
            } else {
                Sleep (1000);
            }
        }
        if (!DeviceIoControl (dev, STRACE_unhook,
                              NULL, 0, NULL, 0, &bytes_out, NULL))
        {
            printf ("ioctl - start failed 0x%x\n", GetLastError ());
            goto exit;
        }
    } else {
        fprintf (stderr, "Failed to load device driver: %d\n",
                 GetLastError ());
    }



 exit:
    if (dev)
        CloseHandle (dev);
    if (hProcess)
        CloseHandle (hProcess);
   /* FIXME: unloading causes page faults currently, so don't unload */
    /* UnloadDeviceDriver (_T ("Strace")); */
}


/*
 * Try to enable the debug privilege
 */
int
enable_debug_priv (void)
{
    HANDLE hToken = 0;
    DWORD dwErr = 0;
    TOKEN_PRIVILEGES newPrivs;

    if (!OpenProcessToken (GetCurrentProcess (),
                           TOKEN_ADJUST_PRIVILEGES,
                           &hToken))
    {
        dwErr = GetLastError ();
        fprintf (stderr, "Unable to open process token: %d\n", dwErr);
        goto exit;
    }

    if (!LookupPrivilegeValue (NULL, SE_DEBUG_NAME,
                               &newPrivs.Privileges[0].Luid))
    {
        dwErr = GetLastError ();
        fprintf (stderr, "Unable to lookup privilege: %d\n", dwErr);
        goto exit;
    }

    newPrivs.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
    newPrivs.PrivilegeCount = 1;
    
    if (!AdjustTokenPrivileges (hToken, FALSE, &newPrivs, 0, NULL, NULL))
    {
        dwErr = GetLastError ();
        fprintf (stderr, "Unable to adjust token privileges: %d\n", dwErr);
        goto exit;
    }

 exit:
    if (hToken)
        CloseHandle (hToken);

    return dwErr;
}
